"""French dialect"""

pyfr = {
    "False": "Faux",
    "None": "Aucun",
    "True": "Vrai",
    "and": "et",
    "as": "comme",
    "assert": "affirmer",
    "async": "async",  # do not translate
    "await": "await",  # as these are not for beginners
    "break": "interrompre",
    "class": "classe",
    "continue": "continuer",
    "def": "définir",
    "del": "supprimer",
    "elif": "sinonsi",
    "else": ["sinon", "pasinterrompu"],
    "except": "siexception",
    "finally": "finalement",
    "for": "pour",
    "from": "de",
    "global": "global",
    "if": "si",
    "import": "importer",
    "in": "dans",
    "is": "est",
    "lambda": "fonction",
    "nonlocal": "nonlocal",
    "not": "pas",
    "or": "ou",
    "pass": "passer",
    "raise": "lever",
    "return": "retourner",
    "try": "essayer",
    "while": "tantque",
    "with": "avec",
    "yield": "cède",
    # Special loop keywords
    "repeat": "répéter",
    "forever": "sansfin",
    "until": "jusquà",
    # a few builtins useful for beginners
    "input": "demander",
    "print": "afficher",
    "range": "intervalle",
    "exit": "quitter",  # useful for console
    # a well-known python expression
    '__name__ == "__main__"': "pasimporté",
}
